// ************************************************************************** //
//
//  BornAgain: simulate and fit scattering at grazing incidence
//
//! @file      GUI/coregui/Models/SessionXML.cpp
//! @brief     Implements reader and writer classes for SessionModel
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2016
//! @authors   Scientific Computing Group at MLZ Garching
//! @authors   Céline Durniak, Marina Ganeva, David Li, Gennady Pospelov
//! @authors   Walter Van Herck, Joachim Wuttke
//
// ************************************************************************** //

#include "ComboProperty.h"
#include "GUIHelpers.h"
#include "GroupItem.h"
#include "ItemFactory.h"
#include "ExternalProperty.h"
#include "GroupItemController.h"
#include "SessionModel.h"
#include "WarningMessageService.h"
#include "SessionItemTags.h"
#include <QtCore/QXmlStreamWriter>

namespace
{
const QString SET_ITEM_PROPERTY_ERROR = "SET_ITEM_PROPERTY_ERROR";
const QString ITEM_IS_NOT_INITIALIZED = "ITEM_IS_NOT_INITIALIZED";

const QString bool_type_name = "bool";
const QString double_type_name = "double";
const QString int_type_name = "int";
const QString qstring_type_name = "QString";

void report_error(WarningMessageService* messageService, SessionItem* item,
                  const QString& error_type, const QString& message);

SessionItem* createItem(SessionItem* item, const QString& modelType, const QString& tag);
}

void SessionXML::writeTo(QXmlStreamWriter* writer, SessionItem* parent)
{
    writer->writeStartElement(parent->model()->getModelTag());
    writer->writeAttribute(SessionXML::ModelNameAttribute, parent->model()->getModelName());

    writeItemAndChildItems(writer, parent);

    writer->writeEndElement(); // m_model_tag
}

void SessionXML::writeItemAndChildItems(QXmlStreamWriter* writer, const SessionItem* item)
{
    if (item->parent()) {
        writer->writeStartElement(SessionXML::ItemTag);
        writer->writeAttribute(SessionXML::ModelTypeAttribute, item->modelType());
        QString tag = item->parent()->tagFromItem(item);
        writer->writeAttribute(SessionXML::TagAttribute, tag);
        writer->writeAttribute(SessionXML::DisplayNameAttribute,
                               item->data(SessionFlags::DisplayNameRole).toString());
        for (int role : item->getRoles()) {
            if (role == Qt::DisplayRole || role == Qt::EditRole)
                SessionXML::writeVariant(writer, item->value(), role);
        }
    }

    for (auto child : item->children())
        writeItemAndChildItems(writer, child);

    if (item->parent())
        writer->writeEndElement(); // ItemTag
}

void SessionXML::writeVariant(QXmlStreamWriter* writer, QVariant variant, int role)
{
    if (variant.isValid()) {
        writer->writeStartElement(SessionXML::ParameterTag);
        QString type_name = variant.typeName();
        writer->writeAttribute(SessionXML::ParameterTypeAttribute, type_name);
        writer->writeAttribute(SessionXML::ParameterRoleAttribute, QString::number(role));

        if (type_name == double_type_name) {
            writer->writeAttribute(SessionXML::ParameterValueAttribute,
                                   QString::number(variant.toDouble(), 'e', 12));
        }

        else if (type_name == int_type_name) {
            writer->writeAttribute(SessionXML::ParameterValueAttribute,
                                   QString::number(variant.toInt()));
        }

        else if (type_name == bool_type_name) {
            writer->writeAttribute(SessionXML::ParameterValueAttribute,
                                   QString::number(variant.toBool()));
        }

        else if (type_name == qstring_type_name) {
            writer->writeAttribute(SessionXML::ParameterValueAttribute, variant.toString());
        }

        else if (type_name == Constants::ExternalPropertyType) {
            ExternalProperty prop = variant.value<ExternalProperty>();
            writer->writeAttribute(SessionXML::ExternalPropertyTextAtt, prop.text());

            QString tcol = prop.color().isValid() ? prop.color().name(QColor::HexArgb) : "";
            writer->writeAttribute(SessionXML::ExternalPropertyColorAtt, tcol);
            writer->writeAttribute(SessionXML::ExternalPropertyIdentifierAtt, prop.identifier());
        }

        else if (type_name == Constants::ComboPropertyType) {
            int currentIndex = variant.value<ComboProperty>().currentIndex();
            writer->writeAttribute(SessionXML::ParameterValueAttribute,
                                   QString::number(currentIndex));
            writer->writeAttribute(SessionXML::ParameterExtAttribute,
                                   variant.value<ComboProperty>().stringOfValues());

        }

        else {
            throw GUIHelpers::Error("SessionModel::writeProperty: Parameter type not supported "
                                    + type_name);
        }

        writer->writeEndElement(); // end ParameterTag
    }
}

void SessionXML::readItems(QXmlStreamReader* reader, SessionItem* parent, QString topTag,
                           WarningMessageService* messageService)
{
    Q_ASSERT(parent);
    const QString start_type = parent->model()->getModelTag();
    while (!reader->atEnd()) {
        reader->readNext();
        if (reader->isStartElement()) {
            if (reader->name() == SessionXML::ItemTag) {
                const QString model_type
                    = reader->attributes().value(SessionXML::ModelTypeAttribute).toString();
                QString tag = reader->attributes().value(SessionXML::TagAttribute).toString();
                QString displayName
                    = reader->attributes().value(SessionXML::DisplayNameAttribute).toString();

                if (!topTag.isEmpty()) {
                    // to handle copying of item to another parent
                    tag = topTag;
                    topTag.clear();
                }
                parent = createItem(parent, model_type, tag);
                parent->setDisplayName(displayName);
            } else if (reader->name() == SessionXML::ParameterTag) {
                SessionXML::readProperty(reader, parent, messageService);
            }
        } else if (reader->isEndElement()) {
            if (reader->name() == SessionXML::ItemTag && parent)
                parent = parent->parent();

            if (reader->name() == start_type)
                break;
        }
    }
}

QString SessionXML::readProperty(QXmlStreamReader* reader, SessionItem* item,
                                 WarningMessageService* messageService)
{
    const QString parameter_name
        = reader->attributes().value(SessionXML::ParameterNameAttribute).toString();
    const QString parameter_type
        = reader->attributes().value(SessionXML::ParameterTypeAttribute).toString();
    const int role = reader->attributes().value(SessionXML::ParameterRoleAttribute).toInt();

    if (!item) {
        QString message
            = QString("Attempt to set property '%1' for non existing item").arg(parameter_name);
        report_error(messageService, item, ITEM_IS_NOT_INITIALIZED, message);
        return parameter_name;
    }

    QVariant variant;
    if (parameter_type == double_type_name) {
        double parameter_value
            = reader->attributes().value(SessionXML::ParameterValueAttribute).toDouble();
        variant = parameter_value;
    }

    else if (parameter_type == int_type_name) {
        int parameter_value
            = reader->attributes().value(SessionXML::ParameterValueAttribute).toInt();
        variant = parameter_value;
    }

    else if (parameter_type == bool_type_name) {
        bool parameter_value
            = reader->attributes().value(SessionXML::ParameterValueAttribute).toInt();
        variant = parameter_value;
    }

    else if (parameter_type == qstring_type_name) {
        QString parameter_value
            = reader->attributes().value(SessionXML::ParameterValueAttribute).toString();
        variant = parameter_value;
    }

    else if (parameter_type == Constants::ExternalPropertyType) {
        QString text = reader->attributes().value(SessionXML::ExternalPropertyTextAtt).toString();
        QString colorName
            = reader->attributes().value(SessionXML::ExternalPropertyColorAtt).toString();
        QString identifier
            = reader->attributes().value(SessionXML::ExternalPropertyIdentifierAtt).toString();

        ExternalProperty property;
        property.setText(text);
        property.setColor(QColor(colorName));
        property.setIdentifier(identifier);
        variant = property.variant();
    }

    else if (parameter_type == Constants::ComboPropertyType) {
        int parameter_value
            = reader->attributes().value(SessionXML::ParameterValueAttribute).toInt();
        QString parameterExt
            = reader->attributes().value(SessionXML::ParameterExtAttribute).toString();

        ComboProperty combo_property;
        combo_property.setStringOfValues(parameterExt);
        combo_property.setCurrentIndex(parameter_value);

        variant = combo_property.variant();
    }

    else {
        throw GUIHelpers::Error("SessionModel::readProperty: "
                                "Parameter type not supported"
                                + parameter_type);
    }

    if (variant.isValid()) {
        item->setData(role, variant);
    }

    return parameter_name;
}

namespace
{
void report_error(WarningMessageService* messageService, SessionItem* item,
                  const QString& error_type, const QString& message)
{
    if (messageService) {
        messageService->send_message(item->model(), error_type, message);
    } else {
        throw GUIHelpers::Error(error_type + QString(" ") + message);
    }
}

SessionItem* createItem(SessionItem* item, const QString& modelType, const QString& tag)
{
    SessionItem* result(nullptr);

    if (item->modelType() == Constants::GroupItemType) {
        if (auto groupItem = dynamic_cast<GroupItem*>(item))
            result = groupItem->getItemOfType(modelType);
    } else {
        if (item->sessionItemTags()->isSingleItemTag(tag))
            result = item->getItem(tag);
        else
            result = item->model()->insertNewItem(modelType, item->index(), -1, tag);
    }
    Q_ASSERT(result);
    return result;
}
}
